/*
**	xdiff -- graphical diff for X
**	read.c: read the files, compute line-starts, parse diff output
**
**	Copyright (C) 1996  Paul J. Lucas
**
**	This program is free software; you can redistribute it and/or modify
**	it under the terms of the GNU General Public License as published by
**	the Free Software Foundation; either version 1, or (at your option)
**	any later version.
** 
**	This program is distributed in the hope that it will be useful,
**	but WITHOUT ANY WARRANTY; without even the implied warranty of
**	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**	GNU General Public License for more details.
** 
**	You should have received a copy of the GNU General Public License
**	along with this program; if not, write to the Free Software
**	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* system */
#include <ctype.h>				/* for isdigit() */
#include <fcntl.h>				/* for open() */
#include <limits.h>				/* for INT_MAX */
#include <stdio.h>
#include <stdlib.h>				/* for exit() */
#include <sys/mman.h>				/* for mmap() */
#include <sys/stat.h>				/* for stat() */
#include <sys/types.h>

/* local */
#include "data.h"
#include "util.h"

#define	INITIAL_LINES	1024
#define	MAX(a,b)	( (a) > (b) ? (a) : (b) )

/* global variable definitions */
line_map		map;

struct diff {
	char	type;				/* 'a', 'd', or 'c' */
	int	line_no[2][2];			/* line[ file ][ begin/end ] */
};

/*************************/
	static int /* bool */
get_diff PJL_ARG_LIST(( d ))
	PJL_ARG_END( register struct diff *d )
{
	/*
	** Parse diff output lines of the form:
	**
	**	n1an3,n4
	**	n1,n2dn3
	**	n1,n2cn3,n4
	**
	** If n1 == n2, n2 is elided; if n3 == n4, n4 is elided.
	*/
	register int c;
	int i, j;

	/*
	** Look for a line that starts with a digit.
	*/
	while ( ( c = getchar() ) != EOF )
		if ( isdigit( c ) )
			goto found_digit;
		else {
			/* ignore the rest of the line */
			while ( ( c = getchar() ) != EOF && c != '\n' )
				;
		}
	return 0 /* false */;

found_digit:
	/*
	** Found a line that starts with a digit -- a diff line.
	*/
	d->line_no[0][0] = d->line_no[0][1] = 0;
	d->line_no[1][0] = d->line_no[1][1] = 0;
	i = j = 0;

	do {
		if ( isdigit( c ) ) {
			d->line_no[i][j] = d->line_no[i][j] * 10 + c - '0';
			continue;
		}
		switch ( c ) {

			case ',':
				j = 1;
				break;

			case 'a':
				/*
				** Adjust the line number diff outputs to be
				** the first line number that is actually
				** different rather than the line - 1 used for
				** an "ed" add command.
				*/
				++d->line_no[0][0];
				/* no break */

			case 'd':
			case 'c':
				d->type = c;
				if ( j == 0 ) {
					/*
					** The range is a single line: copy
					** the low value to the high.
					*/
					d->line_no[0][1] = d->line_no[0][0];
				}
				i = 1;
				j = 0;
				break;

			case '\n':
				if ( d->type == 'd' ) {
					/*
					** Adjust the line number diff outputs
					** to be the first line number that is
					** actually different rather than the
					** line - 1 used for an "ed" delete
					** command.
					*/
					++d->line_no[1][0];
				}
				if ( j == 0 ) {
					/*
					** The range is a single line: copy
					** the low value to the high.
					*/
					d->line_no[1][1] = d->line_no[1][0];
				}
				return 1 /* true */;

			default:
				goto strange_diff;
		}
	} while ( ( c = getchar() ) != EOF );
	return 0 /* false */;

strange_diff:
	fprintf( stderr, "%s: this system has a strange diff\n", me );
	exit( 3 );
}

/*************************/
	static void
grow_map PJL_ARG_LIST(( no_lines, alloc ))
	PJL_ARG_DEF( int no_lines[] )
	PJL_ARG_END( int *alloc )
{
	/*
	** Redetermine the maximum number of lines and grow the vector if
	** necessary.
	*/
	map.max_lines = MAX( no_lines[0], no_lines[1] );
	if ( map.max_lines > *alloc )
		map.set = (line_set*)realloc( map.set,
			(*alloc *= 2) * sizeof( line_set )
		);
}

/*************************/
	static int
linelen PJL_ARG_LIST(( s ))
	PJL_ARG_END( register char const *s )
{
	/*
	** Return the length of a line where '\n' terminates.
	*/
	register char const *c;

	if ( !s )
		return 0;

	for ( c = s; *c && *c != '\n'; ++c )
		;
	return c - s;
}

/*************************/
	static void
map_file PJL_ARG_LIST(( file_name, begin, end ))
	PJL_ARG_DEF( char const *file_name )
	PJL_ARG_DEF( char const **begin )
	PJL_ARG_END( char const **end )
{
	/*
	** Memory-map a file and return begin and end iterators.
	*/
	struct stat stat_buf;
	int fd;

	/* we need to get the file's size */
	if ( stat( file_name, &stat_buf ) == -1 )
		my_perror( file_name );

	if ( ( fd = open( file_name, O_RDONLY ) ) == -1 )
		my_perror( file_name );

	*begin = (char*)mmap( (caddr_t)0, stat_buf.st_size,
		PROT_READ, MAP_SHARED, fd, (off_t)0
	);
	if ( *begin == (char*)-1 )
		my_perror( file_name );

	/* note that this is one past the end */
	*end = *begin + stat_buf.st_size;
}

/*************************/
	int /* bool */
read_files PJL_ARG_LIST(( file_name ))
	PJL_ARG_END( char const *file_name[] )
{
	int differences = 0 /* false */;
	line_info *line[2];
	struct diff diff;
	int no_lines[2];		/* no. lines in files */
	int line_no[2];			/* line number we're on in each file */
	int set_index;
	int map_alloc;			/* how big the map is allocated to be */
	int i, j;

	for ( i = 0; i <= 1; ++i ) {
		int alloc = INITIAL_LINES;
		char const *begin, *end;
		register char const *c;
		char const *line_start;

		map_file( file_name[i], &begin, &end );

		line[i] = (line_info*)malloc( alloc * sizeof( line_info ) );
		line_start = begin;
		no_lines[i] = 0;

		/*
		** Scan through the file to find out where all the lines start.
		*/
		for ( c = begin; c != end; ++c )
			if ( *c == '\n' ) {
				line_info *const p = &line[ i ][ no_lines[i] ];
				p->length = linelen( p->text = line_start );
				line_start = c + 1;

				/* see if we've run out of line storage */
				if ( ++no_lines[i] >= alloc )
					line[i] = (line_info*)realloc( line[i],
						(alloc *= 2)
						* sizeof( line_info )
					);
			}
	}

	/*
	** The number of lines in the display is the maximum of the number of
	** lines in the two files.
	*/
	map.max_lines = MAX( no_lines[0], no_lines[1] );

	map_alloc = map.max_lines;
	map.set = (line_set*)malloc( map_alloc * sizeof( line_set ) );
	line_no[0] = line_no[1] = set_index = 0;

	/*
	** Process each diff in turn.
	*/
	while ( get_diff( &diff ) ) {
		int spill_over = 0 /* false */;
		int tmp;

		/* there's at least one difference */
		differences = 1 /* true */;
#ifdef	DEBUG
		printf( "read: diff=%d,%d%c%d,%d\n",
			diff.line_no[0][0],
			diff.line_no[0][1],
			diff.type,
			diff.line_no[1][0],
			diff.line_no[1][1]
		);
#endif
		/*
		** Mark all the lines as the same up to the current diff.
		*/
		tmp = set_index;
		for ( j = 0; j <= 1; ++j ) {
			int limit = diff.line_no[j][0] - 1;
#ifdef	DEBUG
			printf( "read: catching [%d] up to line: %d\n",
				j, limit
			);
#endif
			set_index = tmp;
			while ( line_no[j] < limit ) {
				line[j][ line_no[j] ].type = same_line;
				map.set[ set_index ].line[j] =
					&line[j][ line_no[j]++ ];
				++set_index;
			}
		}

		switch ( diff.type ) {

			case 'c': {
				int limit[2];
				int dd[2];	/* diff line_no difference */
				int tmp;

				/*
				** Changes don't have to have a 1-to-1 mapping
				** between lines.  N lines can change into N+M
				** lines.  The N lines will be marked as
				** changed, but the M lines will be marked as
				** new by falling through into the next case.
				*/
				limit[0] = diff.line_no[0][1];
				limit[1] = diff.line_no[1][1];
				dd[0] = limit[0] - diff.line_no[0][0];
				dd[1] = limit[1] - diff.line_no[1][0];
				for ( j = 0; j <= 1; ++j )
					if ( dd[j] > dd[!j] ) {
						/*
						** We must increase the number
						** of lines in one of the files
						** by the difference in the
						** number of lines in the diff
						** to create a "hole" in the
						** file so that, when the files
						** are displayed, the lines
						** that should line up continue
						** to do so.
						*/
						no_lines[!j] += dd[j] - dd[!j];
#ifdef	DEBUG
						printf( "read: [%d] += %d\n",
							!j, dd[j] - dd[!j]
						);
#endif
					}

				/*
				** Since the number of lines in one of the two
				** files may have just changed, we must
				** redetermine the maximum number of lines and
				** grow the vector if necessary.
				*/
				grow_map( no_lines, &map_alloc );

				/*
				** Assuming the change is not 1-to-1, alter the
				** diff type to mark the proper file's extra
				** lines appropriately if/when we fall through
				** to the next case.
				*/
				j = dd[0] > dd[1];
				diff.type = "ad"[j];

				/*
				** Also, only go up to the limit of the smaller
				** number of lines that are changed; the
				** remainder will be handled by the next case.
				*/
				limit[!j] = diff.line_no[!j][0] + dd[j];

				/*
				** For both files, mark the lines as different.
				*/
				tmp = set_index;
				for ( j = 0; j <= 1; ++j ) {
					set_index = tmp;
					while ( line_no[j] < limit[j] ) {
						line[j][ line_no[j] ].type =
							diff_line;
						map.set[ set_index++ ].line[j] =
							&line[j][ line_no[j] ];
						++line_no[j];
					}
				}
				if ( dd[0] == dd[1] )
					break;
				/*
				** The change is not 1-to-1: Spill over into
				** the next case to mark the extra lines as
				** new.
				*/
				spill_over = 1 /* true */;
				/* no break */
			}

			case 'a':
			case 'd':
				j = diff.type == 'a';

				if ( !spill_over ) {
					/*
					** We must increase the number of lines
					** in one of the files by the
					** difference in the number of lines in
					** the diff to create a "hole" in the
					** file so that, when the files are
					** displayed, the lines that should
					** line up continue to do so.  (If we
					** spilled over from the Change case,
					** that code already took care of
					** increasing the number of lines.)
					*/
					no_lines[!j] += diff.line_no[j][1] -
						diff.line_no[j][0] + 1;
#ifdef	DEBUG
					printf( "read: [%d] += %d\n", !j,
						diff.line_no[j][1] -
						diff.line_no[j][0] + 1
					);
#endif
					/*
					** Since the number of lines in one of
					** the two files just changed, we must
					** redetermine the maximum number of
					** lines and grow the vector if
					** necessary.
					*/
					grow_map( no_lines, &map_alloc );
				}

				/*
				** Mark all the lines in the diff as new in one
				** file and set the line pointer to null in the
				** other file to create a "hole."
				*/
				while ( line_no[j] < diff.line_no[j][1] ) {
					line[j][ line_no[j] ].type = new_line;
					map.set[ set_index ].line[ j] =
						&line[j][ line_no[j] ];
					map.set[ set_index ].line[!j] = 0;
					++set_index;
					++line_no[j];
				}
				break;
		}
	}
#ifdef	DEBUG
	printf( "read: set_index=%d max_lines=%d\n", set_index, map.max_lines );
#endif

	/*
	** There are no more differences -- finish mapping of remaining lines,
	** if any, marking them as the same.
	*/
	while ( set_index < map.max_lines ) {
		line[0][ line_no[0] ].type = same_line;
		line[1][ line_no[1] ].type = same_line;
		map.set[ set_index ].line[0] = &line[0][ line_no[0]++ ];
		map.set[ set_index ].line[1] = &line[1][ line_no[1]++ ];
		++set_index;
	}

	return differences;
}
